// <copyright file="Image_MultiTrack.cs" company="QutEcoacoustics">
// All code in this file and all associated files are the copyright and property of the QUT Ecoacoustics Research Group (formerly MQUTeR, and formerly QUT Bioacoustics Research Group).
// </copyright>

namespace AudioAnalysisTools.StandardSpectrograms
{
    using System;
    using System.Collections.Generic;
    using AudioAnalysisTools.Events.Drawing;
    using AudioAnalysisTools.Events.Interfaces;
    using AudioAnalysisTools.Events.Tracks;
    using SixLabors.ImageSharp;
    using SixLabors.ImageSharp.Drawing.Processing;
    using SixLabors.ImageSharp.PixelFormats;
    using SixLabors.ImageSharp.Processing;
    using TowseyLibrary;

    public class Image_MultiTrack : IDisposable
    {
        private readonly List<ImageTrack> tracks = new List<ImageTrack>();
        private double superImposedMaxScore;
        private int[] freqHits;
        private int nyquistFreq; //sets the frequency scale for drawing events
        private int freqBinCount;
        private double freqBinWidth;
        private double framesPerSecond;

        /// <summary>
        /// Initializes a new instance of the <see cref="Image_MultiTrack"/> class.
        /// CONSTRUCTOR.
        /// </summary>
        public Image_MultiTrack(Image<Rgb24> image)
        {
            this.SonogramImage = image;
            this.Points = new List<PointOfInterest>();
        }

        public Image<Rgb24> SonogramImage { get; private set; }

        public IEnumerable<ImageTrack> Tracks => this.tracks;

        public IEnumerable<AcousticEvent> EventList { get; set; }

        public List<Track> SpectralTracks { get; set; }

        public double[,] SuperimposedMatrix { get; set; }

        public double[,] SuperimposedRedTransparency { get; set; }

        public double[,] SuperimposedRainbowTransparency { get; set; }

        public int[,] SuperimposedDiscreteColorMatrix { get; set; }

        //private Point[] points;
        public List<PointOfInterest> Points { get; set; }

        public void AddTrack(ImageTrack track)
        {
            this.tracks.Add(track);
        }

        public void AddEvents(IEnumerable<AcousticEvent> list, int nyquist, int freqBinCount, double framesPerSecond)
        {
            this.EventList = list;
            this.nyquistFreq = nyquist;
            this.freqBinCount = freqBinCount;
            this.framesPerSecond = framesPerSecond;
            this.freqBinWidth = nyquist / (double)freqBinCount;
        }

        public void AddPoints(IEnumerable<PointOfInterest> points)
        {
            // this.points.AddRange(points);

            // scan for preexisting coordinates
            foreach (var pointOfInterest in points)
            {
                // copied to satisfy closure constraint
                PointOfInterest localCopy = pointOfInterest;

                // search current points to see if any share the same coordinates

                var match = this.Points.IndexOf(poi => poi.Point == localCopy.Point);
                if (match >= 0)
                {
                    // if they do share the same coordinates, overwrite the old one
                    this.Points[match] = pointOfInterest;
                }
                else
                {
                    // otherwise, add new point to list
                    this.Points.Add(pointOfInterest);
                }
            }
        }

        public void OverlayRedMatrix(double[,] m, double maxScore)
        {
            //this.SuperimposedMatrix = m; // TODO:  This line does not work !!?? Use next line
            this.SuperimposedRedTransparency = m;
            this.superImposedMaxScore = maxScore;
        }

        public void OverlayRedTransparency(double[,] m)
        {
            this.SuperimposedRedTransparency = m;
        }

        public void OverlayRainbowTransparency(double[,] m)
        {
            this.SuperimposedRainbowTransparency = m;
        }

        public void OverlayDiscreteColorMatrix(int[,] m)
        {
            this.SuperimposedDiscreteColorMatrix = m;
        }

        public void AddFreqHitValues(int[] f, int nyquist)
        {
            this.freqHits = f;
            this.nyquistFreq = nyquist;
        }

        public void AddTracks(List<Track> tracks, double framesPerSecond, double freqBinWidth)
        {
            this.freqBinWidth = freqBinWidth;
            this.framesPerSecond = framesPerSecond;
            this.SpectralTracks = tracks;
        }

        /// <summary>
        /// WARNING: This method calls Image_MultiTrack.GetImage().
        /// In some circumstances GetImage() cannot manage images with an area larger than 10,385,000 pixels.
        /// This means it cannot handle recording sonograms longer than 2 minutes.
        /// </summary>
        public void Save(string path)
        {
            var image = this.GetImage();
            if (image == null)
            {
                Log.WriteLine("MultiTrackImage.Save() - WARNING: NULL IMAGE.Cannot save to: " + path);
                return;
            }

            image.Save(path);
        }

        /// <summary>
        /// WARNING: graphics.DrawImage() or GDI cannot draw an image that is too big, typically
        /// with an area larger than 10,385,000 pixels (Jiro estimated > 40000 pixels).
        /// This means it cannot handle recording sonograms longer than 2 minutes.
        /// Therefore call a recursive method to draw the image.
        /// </summary>
        public Image<Rgb24> GetImage()
        {
            // Calculate total height of the bmp
            var height = this.CalculateImageHeight();

            // set up a new image having the correct dimensions
            var imageToReturn = new Image<Rgb24>(this.SonogramImage.Width, height);

            // need to do this before get Graphics because cannot PutPixels into Graphics object.
            if (this.SuperimposedRedTransparency != null)
            {
                this.SonogramImage = this.OverlayRedTransparency((Image<Rgb24>)this.SonogramImage);
            }

            if (this.SuperimposedMatrix != null)
            {
                this.SonogramImage = this.OverlayMatrix((Image<Rgb24>)this.SonogramImage);
            }

            // create new graphics canvas and add in the sonogram image
            imageToReturn.Mutate(g =>
            {
                g.DrawImage(this.SonogramImage, 1f);

                // draw events first because their rectangles can cover other features
                if (this.EventList != null)
                {
                    var hitImage = new Image<Rgb24>(imageToReturn.Width, height);

                    //hitImage.MakeTransparent();
                    foreach (AcousticEvent e in this.EventList)
                    {
                        // TODO TODO Fix this call. I am having problem with managing SixLabors.ImageSharp
                        e.DrawEvent(hitImage, this.framesPerSecond, this.freqBinWidth, this.SonogramImage.Height);
                    }

                    g.DrawImage(hitImage, 0, 0);
                }

                // draw events first because their rectangles can cover other features
                if (this.Points != null)
                {
                    // var stats = new StatDescriptive(this.points.Select(p => p.Item2).ToArray());
                    // stats.Analyze();
                    foreach (PointOfInterest poi in this.Points)
                    {
                        //poi.DrawPoint(g, this.Points, this.sonogramImage.Height);
                        const int radius = 12;
                        poi.DrawBox(g, this.Points, radius);
                    }
                }

                if (this.SpectralTracks != null)
                {
                    var converter = new UnitConverters(
                        segmentStartOffset: 0.0,
                        segmentDuration: this.framesPerSecond * this.SonogramImage.Width,
                        nyquistFrequency: this.freqBinWidth * this.SonogramImage.Height,
                        imageWidth: this.SonogramImage.Width,
                        imageHeight: this.SonogramImage.Height);

                    var renderingOptions = new EventRenderingOptions(converter);

                    foreach (var t in this.SpectralTracks)
                    {
                        t.Draw(g, renderingOptions);
                    }
                }

                if (this.freqHits != null)
                {
                    this.DrawFreqHits(g);
                }

                if (this.SuperimposedRainbowTransparency != null)
                {
                    this.OverlayRainbowTransparency(g, (Image<Rgb24>)this.SonogramImage);
                }

                if (this.SuperimposedDiscreteColorMatrix != null)
                {
                    this.OverlayDiscreteColorMatrix(g);
                }
            });

            // now add tracks to the image
            int offset = this.SonogramImage.Height;
            foreach (ImageTrack track in this.tracks)
            {
                track.topOffset = offset;
                track.bottomOffset = offset + track.Height - 1;
                track.DrawTrack(imageToReturn.CloneAs<Rgb24>());
                offset += track.Height;
            }

            return imageToReturn;
        }

        private int CalculateImageHeight()
        {
            int totalHeight = this.SonogramImage.Height;
            foreach (ImageTrack track in this.tracks)
            {
                totalHeight += track.Height;
            }

            return totalHeight;
        }

        public void DrawFreqHits(IImageProcessingContext g)
        {
            int length = this.freqHits.Length;
            Pen p1 = new Pen(Color.Red, 1);

            for (int x = 0; x < length; x++)
            {
                if (this.freqHits[x] <= 0)
                {
                    continue;
                }

                int y = (int)(this.SonogramImage.Height * (1 - (this.freqHits[x] / (double)this.nyquistFreq)));

                //g.DrawRectangle(p1, x, y, x + 1, y + 1);
                g.DrawLine(p1, x, y, x, y + 1);

                // g.DrawTextSafe(e.Name, Drawing.Tahoma6, Color.Black, new PointF(x, y - 1));
            }
        }

        /// <summary>
        /// superimposes a matrix of scores on top of a sonogram.
        /// Only draws lines on every second row so that the underling sonogram can be discerned.
        /// </summary>
        public void OverlayMatrix(IImageProcessingContext g)
        {
            //int paletteSize = 256;
            var pens = ImageTools.GetRedGradientPalette(); //size = 256

            int rows = this.SuperimposedMatrix.GetLength(0);
            int cols = this.SuperimposedMatrix.GetLength(1);
            int imageHt = this.SonogramImage.Height - 1; //subtract 1 because indices start at zero

            //ImageTools.DrawMatrix(DataTools.MatrixRotate90Anticlockwise(this.SuperimposedMatrix), @"C:\SensorNetworks\WavFiles\SpeciesRichness\Dev1\superimposed1.png", false);

            // traverse columns - skip DC column
            for (int c = 1; c < cols; c++)
            {
                for (int r = 0; r < rows; r++)
                {
                    if (this.SuperimposedMatrix[r, c] == 0.0)
                    {
                        continue;
                    }

                    double normScore = this.SuperimposedMatrix[r, c] / this.superImposedMaxScore;

                    //int penID = (int)(paletteSize * normScore);
                    //if (penID >= paletteSize) penID = paletteSize - 1;
                    var brush = new SolidBrush(Color.Red);
                    g.FillRectangle(brush, r, imageHt - c, 1, 1); //THIS DRAWS A PIXEL !!!!
                }

                //c++; //only draw on every second row.
            }
        }

        /// <summary>
        /// superimposes a matrix of scores on top of a sonogram.
        /// Only draws lines on every second row so that the underling sonogram can be discerned.
        /// </summary>
        public Image<Rgb24> OverlayMatrix(Image<Rgb24> bmp)
        {
            Image<Rgb24> newBmp = (Image<Rgb24>)bmp.Clone();

            //int paletteSize = 256;
            var pens = ImageTools.GetRedGradientPalette(); //size = 256

            int rows = this.SuperimposedMatrix.GetLength(0);
            int cols = this.SuperimposedMatrix.GetLength(1);
            int imageHt = this.SonogramImage.Height - 1; //subtract 1 because indices start at zero

            // traverse columns - skip DC column
            for (int c = 1; c < cols; c++)
            {
                for (int r = 0; r < rows; r++)
                {
                    if (this.SuperimposedMatrix[r, c] == 0.0)
                    {
                        continue;
                    }

                    double normScore = this.SuperimposedMatrix[r, c] / this.superImposedMaxScore;

                    //int penID = (int)(paletteSize * normScore);
                    //if (penID >= paletteSize) penID = paletteSize - 1;
                    //g.DrawLine(pens[penID], r, imageHt - c, r + 1, imageHt - c);
                    //g.DrawLine(new Pen(Color.Red, 1), r, imageHt - c, r + 1, imageHt - c);
                    newBmp[r, imageHt - c] = Color.Red;
                }
            }

            return newBmp;
        } //OverlayMatrix()

        /// <summary>
        /// superimposes a matrix of scores on top of a sonogram.
        /// </summary>
        public Image<Rgb24> OverlayRedTransparency(Image<Rgb24> bmp)
        {
            return Image_MultiTrack.OverlayScoresAsRedTransparency(bmp, this.SuperimposedRedTransparency);
        }

        /// <summary>
        /// It is assumed that the spectrogram image is grey scale.
        /// NOTE: The score matrix must consist of reals in [0.0, 1.0].
        /// NOTE: The image and the score matrix must have the same number of rows and columns.
        /// In case of a spectrogram, it is assumed that the rows are frequency bins and the columns are individual spectra.
        /// </summary>
        /// <param name="bmp">the spectrogram image.</param>
        /// <param name="hits">the matrix of scores or hits.</param>
        public static Image<Rgb24> OverlayScoresAsRedTransparency(Image<Rgb24> bmp, double[,] hits)
        {
            Image<Rgb24> newBmp = (Image<Rgb24>)bmp.Clone();
            int rows = hits.GetLength(0);
            int cols = hits.GetLength(1);

            if (rows != bmp.Height || cols != bmp.Width)
            {
                LoggedConsole.WriteErrorLine("ERROR: Image and hits matrix do not have the same dimensions.");
                return bmp;
            }

            for (int r = 0; r < rows; r++)
            {
                for (int c = 0; c < cols; c++)
                {
                    if (hits[r, c] <= 0.0)
                    {
                        continue;
                    }

                    if (hits[r, c] > 1.0)
                    {
                        hits[r, c] = 1.0;
                    }

                    var pixel = bmp[r, c];
                    byte value = (byte)Math.Floor(hits[r, c] * 255);
                    newBmp[r, c] = Color.FromRgb(value, pixel.G, pixel.B);
                }
            }

            return newBmp;
        }

        /// <summary>
        /// Overlays a matrix of scores on an image, typically a spectrogram image.
        /// It is assumed that the spectrogram image is grey scale.
        /// NOTE: The score matrix must consist of integers from 0 to 255.
        /// NOTE: The image and the score matrix must have the same number of rows and columns.
        /// In case of a spectrogram, it is assumed that the rows are frequency bins and the columns are individual spectra.
        /// </summary>
        /// <param name="bmp">the spectrogram image.</param>
        /// <param name="hits">the matrix of scores or hits.</param>
        /// <returns>The new image with overlay of scores as red transparency.</returns>
        public static Image<Rgb24> OverlayScoresAsRedTransparency(Image<Rgb24> bmp, int[,] hits)
        {
            Image<Rgb24> newBmp = (Image<Rgb24>)bmp.Clone();
            int rows = hits.GetLength(0);
            int cols = hits.GetLength(1);

            if (rows != bmp.Height || cols != bmp.Width)
            {
                LoggedConsole.WriteErrorLine("ERROR: Image and hits matrix do not have the same dimensions.");
                return bmp;
            }

            for (int r = 0; r < rows; r++)
            {
                for (int c = 0; c < cols; c++)
                {
                    if (hits[r, c] <= 0)
                    {
                        continue;
                    }

                    if (hits[r, c] > 255)
                    {
                        hits[r, c] = 255;
                    }

                    var pixel = bmp[c, r];
                    newBmp[c, r] = Color.FromRgb((byte)hits[r, c], pixel.G, pixel.B);
                }
            }

            return newBmp;
        }

        /// <summary>
        /// WARNING: THis method is yet to be debugged and tested following move to SixLabors drawing libraries.
        /// superimposes a matrix of scores on top of a sonogram.
        /// USES A PALLETTE of ten RAINBOW colours.
        /// ASSUME MATRIX is NORMALIZED IN [0,1].
        /// </summary>
        public void OverlayRainbowTransparency(IImageProcessingContext g, Image<Rgb24> bmp)
        {
            Color[] palette = { Color.Crimson, Color.Red, Color.Orange, Color.Yellow, Color.Lime, Color.Green, Color.Blue, Color.Indigo, Color.Violet, Color.Purple };
            int rows = this.SuperimposedRainbowTransparency.GetLength(0);
            int cols = this.SuperimposedRainbowTransparency.GetLength(1);
            int imageHt = this.SonogramImage.Height - 1; //subtract 1 because indices start at zero

            for (int r = 0; r < rows; r++)
            {
                // traverse columns - skip DC column
                for (int c = 1; c < cols; c++)
                {
                    double value = this.SuperimposedRainbowTransparency[r, c];
                    if (value <= 0.0)
                    {
                        continue; //nothing to show
                    }

                    var pixel = bmp[r, imageHt - c];
                    if (pixel.R > 250)
                    {
                        continue; //by-pass white
                    }

                    int index = (int)Math.Floor(value * 9); // get index into pallette
                    if (index > 9)
                    {
                        index = 9;
                    }

                    var newColor = palette[index].ToPixel<Rgb24>();
                    double factor = pixel.R / (255 * 1.2);  // 1.2 is a color intensity adjustment
                    var red = (byte)Math.Floor(newColor.R + ((255 - newColor.R) * factor));
                    var grn = (byte)Math.Floor(newColor.G + ((255 - newColor.G) * factor));
                    var blu = (byte)Math.Floor(newColor.B + ((255 - newColor.B) * factor));
                    g.DrawLine(new Pen(Color.FromRgb(red, grn, blu), 1), r, imageHt - c, r + 1, imageHt - c);
                    c++; //every second column
                }
            }
        } //OverlayRainbowTransparency()

        /// <summary>
        /// superimposes a matrix of scores on top of a sonogram. USES RAINBOW PALLETTE.
        /// ASSUME MATRIX consists of integers >=0.
        /// </summary>
        private void OverlayDiscreteColorMatrix(IImageProcessingContext g)
        {
            int rows = this.SuperimposedDiscreteColorMatrix.GetLength(0);
            int cols = this.SuperimposedDiscreteColorMatrix.GetLength(1);
            MatrixTools.MinMax(this.SuperimposedDiscreteColorMatrix, out var min, out var max);
            int palleteLength = ImageTools.DarkColors.Length;

            //Color[] palette = { Color.Crimson, Color.Red, Color.Orange, Color.Yellow, Color.Lime, Color.Green, Color.Blue, Color.Indigo, Color.Violet, Color.Purple };
            int imageHt = this.SonogramImage.Height - 1; //subtract 1 because indices start at zero

            for (int r = 0; r < rows; r++)
            {
                //traverse columns - skip DC column
                for (int c = 1; c < cols; c++)
                {
                    int index = this.SuperimposedDiscreteColorMatrix[r, c];
                    if (index <= 0)
                    {
                        continue; //nothing to show
                    }

                    // Color pixel = bmp.GetPixel(r, imageHt - c);
                    // if (pixel.R > 250) continue; //by-pass white
                    // int index = (int)Math.Floor((value * 9));//get index into pallette
                    if (index >= palleteLength)
                    {
                        index = index % palleteLength;
                    }

                    var newColor = ImageTools.DarkColors[index];

                    //double factor = pixel.R / (double)(255 * 1.2);  //1.2 is a color intensity adjustment
                    //int red = (int)Math.Floor(newColor.R + ((255 - newColor.R) * factor));
                    //int grn = (int)Math.Floor(newColor.G + ((255 - newColor.G) * factor));
                    //int blu = (int)Math.Floor(newColor.B + ((255 - newColor.B) * factor));
                    //g.DrawLine(new Pen(Color.FromRgb(red, grn, blu)), r, imageHt - c, r + 1, imageHt - c);
                    g.DrawLine(new Pen(newColor, 1), r, imageHt - c, r + 1, imageHt - c);
                }
            }
        } //OverlayDiscreteColorMatrix()

        public void Dispose()
        {
            this.EventList = null;
            this.SonogramImage.Dispose();
        }
    }
}